use std::{fs::File, io::Write, path::PathBuf};

use radicle_ci_broker::{
    broker_event::{BrokerEvent, BrokerEvents},
    event::NodeEventSource,
};

use super::*;

/// Listen on node events from the local Radicle node and write them
/// to the standard output as JSON. If requested, also write the
/// corresponding broker events to the specified file as JSON, one per
/// line.
#[derive(Parser)]
pub struct ListenNode {
    /// Write broker events to this file as JSON Lines.
    #[clap(long)]
    output: Option<PathBuf>,

    /// Write node events to this file as JSON Lines.
    #[clap(long)]
    node_events: Option<PathBuf>,
}

impl Leaf for ListenNode {
    fn run(&self, _args: &Args) -> Result<(), CibToolError> {
        let mut output: Option<File> = if let Some(filename) = &self.output {
            println!("writing broker events to {}", filename.display());
            Some(
                File::create(filename)
                    .map_err(|e| CibToolError::OpenEventFile(filename.into(), e))?,
            )
        } else {
            None
        };

        let mut node_events: Option<File> = if let Some(filename) = &self.node_events {
            println!("writing node events to {}", filename.display());
            Some(
                File::create(filename)
                    .map_err(|e| CibToolError::OpenEventFile(filename.into(), e))?,
            )
        } else {
            None
        };

        let profile = util::load_profile()?;
        let mut source =
            NodeEventSource::new(&profile).map_err(CibToolError::CreateNodeEventSource)?;

        loop {
            eprintln!("waiting for events from node");
            let result = source.node_event().map_err(CibToolError::GetNodeEvent);
            match result {
                Ok(Some(event)) => {
                    println!("got node event: {event:?}");
                    if let Some(file) = &mut node_events {
                        let json = serde_json::to_string(&event)
                            .map_err(|e| CibToolError::NodeEventToJson(event.clone(), e))?;
                        file.write_all(json.as_bytes()).unwrap();
                    }

                    if let Some(output) = &mut output {
                        let mut n = 0;
                        for event in BrokerEvent::from_node_event(&event)
                            .map_err(|e| CibToolError::BrokerEvent(event.clone(), e))?
                        {
                            let json = serde_json::to_string(&event).map_err(|e| {
                                CibToolError::NewBrokerEventToJson(event.clone(), e)
                            })?;
                            output
                                .write_all(format!("{}\n", json).as_bytes())
                                .map_err(CibToolError::WriteEvent)?;
                            n += 1;
                        }
                        println!("created {n} broker events from node event");
                    }
                }
                Ok(None) => {
                    eprintln!("EOF");
                    break;
                }
                Err(err) => {
                    eprintln!("ERROR: {err}");
                    break;
                }
            }
        }
        Ok(())
    }
}

/// Listen on node events from the local Radicle node.
#[derive(Parser)]
pub struct FilterNode {
    /// Read filter from this file.
    #[clap(long)]
    filter: PathBuf,

    /// Read broker events from this file.
    #[clap(long)]
    events: PathBuf,
}

impl Leaf for FilterNode {
    fn run(&self, _args: &Args) -> Result<(), CibToolError> {
        let filter = radicle_ci_broker::filter::EventFilter::from_file(&self.filter)
            .map_err(|e| CibToolError::LoadEventFilters(self.filter.clone(), e))?;
        let json = serde_json::to_string_pretty(&filter)
            .map_err(|e| CibToolError::EventFilterToJson(filter.clone(), e))?;
        print!("Event filter:\n{}", indent(&json, 4));

        let events =
            BrokerEvents::from_file(&self.events).map_err(CibToolError::LoadBrokerEvents)?;

        for event in events.iter() {
            let json = serde_json::to_string_pretty(&filter)
                .map_err(|e| CibToolError::NewBrokerEventToJson(event.clone(), e))?;
            if filter.allows(event) {
                print!("Allowed:\n{}", indent(&json, 4));
            } else {
                print!("Denied:\n{}", indent(&json, 4));
            }
        }
        Ok(())
    }
}

fn indent(text: &str, indent: usize) -> String {
    let mut s = String::new();
    for line in text.lines() {
        for _ in 0..indent {
            s.push(' ');
        }
        s.push_str(line);
        s.push('\n');
    }
    s
}
